package de.moritzruth.dracula_musical.device

import de.moritzruth.dracula_musical.device.StairvilleSplb.Effect
import de.moritzruth.theaterdsl.device.*
import de.moritzruth.theaterdsl.dmx.DmxAddress
import de.moritzruth.theaterdsl.dmx.DmxDataWriter
import de.moritzruth.theaterdsl.value.Color
import de.moritzruth.theaterdsl.value.Percentage
import de.moritzruth.theaterdsl.value.degrees
import de.moritzruth.theaterdsl.value.percent
import kotlinx.collections.immutable.ImmutableSet
import kotlinx.collections.immutable.persistentSetOf
import kotlin.math.PI
import kotlin.math.min
import kotlin.math.roundToLong
import kotlin.math.sin

class StairvilleTlbButItsSplb(override val firstChannel: DmxAddress, val inwardsIsRight: Boolean = false) : Device {
    companion object {
        private const val NUMBER_OF_SEGMENTS = 9
    }

    override val numberOfChannels: UInt = 27u

    val brightness = PercentageDV()
    val strobeFrequency = PercentageDV()

    val color = ColorDV(Color.WHITE.copy(brightness = 0.percent))
    val white = PercentageDV()

    val effect = object : EffectDV<Effect?, Array<Color>>(null) {
        override fun compute(time: Long, effect: Effect?): Array<Color> {
            val white = white.getCurrentValue()
            val color = if (white == 0.percent) color.getCurrentValue() else Color(30.degrees, 85.percent, white)

            fun fadeInOut(
                time: Long,
                rampUpDuration: Long,
                sustainDuration: Long,
                rampDownDuration: Long,
                peak: Percentage = 100.percent,
                start: Percentage = 0.percent,
                end: Percentage = 0.percent
            ): Percentage {
                return if (time < rampUpDuration) {
                    val progress = time / rampUpDuration.toDouble()
                    val delta = peak.value - start.value
                    Percentage(progress * delta + start.value)
                } else if (time < rampUpDuration + sustainDuration) {
                    peak
                } else {
                    val progress = min((time - rampUpDuration - sustainDuration) / rampDownDuration.toDouble(), 1.0)
                    val delta = peak.value - end.value
                    Percentage(peak.value - progress * delta)
                }
            }

            fun rIndex(condition: Boolean, index: Int) = if (condition) NUMBER_OF_SEGMENTS - index else index

            return when (effect) {
                null -> Array(NUMBER_OF_SEGMENTS) { color }
                Effect.HECTIC_SWITCHING -> {
                    val step = (time / 100).mod(2)
                    Array(NUMBER_OF_SEGMENTS) { index -> if ((index + step) % 2 == 0) color else Color.BLACK }
                }

                Effect.THEATRE_SWITCHING -> {
                    val interval = 500L
                    Array(NUMBER_OF_SEGMENTS) { index ->
                        color.multiplyBrightness(
                            fadeInOut(
                                (time + (if (index % 2 == 0) interval / 2 else 0)).mod(interval),
                                rampUpDuration = (interval * (1 / 8.0)).roundToLong(),
                                sustainDuration = (interval * (1 / 4.0)).roundToLong(),
                                rampDownDuration = (interval * (1 / 4.0)).roundToLong(),
                            )
                        )
                    }
                }

                Effect.FLOW_INWARDS, Effect.FLOW_INWARDS_LONG, Effect.FLOW_OUTWARDS, Effect.FLOW_OUTWARDS_LONG -> {
                    val isOutwards = effect == Effect.FLOW_OUTWARDS || effect == Effect.FLOW_OUTWARDS_LONG
                    val isLong = effect == Effect.FLOW_INWARDS_LONG || effect == Effect.FLOW_OUTWARDS_LONG
                    val interval = 1000.0
                    val scale = if (isLong) 1.5 else 0.5

                    Array(NUMBER_OF_SEGMENTS) { index ->
                        val value = Percentage(
                            sin((time + rIndex(inwardsIsRight xor isOutwards, index) * (interval / NUMBER_OF_SEGMENTS)) / interval * 2 * PI / scale)
                                    / 2 + 0.5
                        )

                        color.multiplyBrightness(value)
                    }
                }

                Effect.BUMP_INWARDS -> {
                    val interval = 350.0

                    Array(NUMBER_OF_SEGMENTS) { index ->
                        color.multiplyBrightness(
                            fadeInOut(
                                (time + rIndex(inwardsIsRight, index) * (interval / NUMBER_OF_SEGMENTS)).mod(interval).roundToLong(),
                                rampUpDuration = 0,
                                sustainDuration = 0,
                                rampDownDuration = (interval * 0.75).roundToLong(),
                            )
                        )
                    }
                }

                Effect.CHASE_BACK_AND_FORTH -> {
                    val interval = 350
                    val fullInterval = interval * 2
                    val isMovingNegative = time.mod(fullInterval) < interval
                    val frontIndex = (if (isMovingNegative) time.mod(interval) else interval - time.mod(interval)) / interval.toDouble() * NUMBER_OF_SEGMENTS

                    Array(NUMBER_OF_SEGMENTS) { index ->
                        val t = fullInterval + if (isMovingNegative) ((index - frontIndex) * (interval.toDouble() / NUMBER_OF_SEGMENTS))
                        else ((frontIndex - index) * (interval.toDouble() / NUMBER_OF_SEGMENTS))

                        color.multiplyBrightness(
                            fadeInOut(
                                t.roundToLong(),
                                rampUpDuration = 0,
                                sustainDuration = 0,
                                rampDownDuration = (interval * (3 / 8.0)).roundToLong(),
                            )
                        )
                    }
                }

                else -> TODO()
            }
        }
    }

    override val dvs: ImmutableSet<DynamicValue<*>> = persistentSetOf(
        brightness,
        strobeFrequency,
        color,
        white,
        effect,
    )

    override fun writeDmxData(writer: DmxDataWriter, isLightBehindCurtainOn: Boolean) {
        val brightness = brightness.getCurrentValue()
        val segments = effect.getCurrentValue()

        for (segment in segments) {
            val (r, g, b) = segment.multiplyBrightness(brightness).powBrightness(2.5).getRGB()
            writer.writePercentage(r)
            writer.writePercentage(g)
            writer.writePercentage(b)
        }
    }
}